# Sherlock version 1.0.0
# SherlockEngine.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Sherlock, a primitive solver for numerical contraint satisfaction problems.
# Sherlock is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Sherlock is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Sherlock.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Engine

# 1) Imported modules
# -------------------

import string
import sys
import re
import math
import time

from SherlockModels import *
from SherlockParser import *
from SherlockPrinter import *
from SherlockSolver import *

# 2) Engine
# ---------

class SherlockEngine:
  def __init__(self):
    self._model = SherlockModel()
    self._parser = SherlockParser()
    self._printer = SherlockPrinter()
    self._solver = SherlockSolver()
    self._problem = None

  def LoadAndSolveProblem(self, inputFileName, problemName, algorithmName, outputFileName):
    error = self._parser.ImportFile(self._model, inputFileName)
    if error!=0:
      return
    self._problem = self._model.GetProblem(problemName)
    if self._problem==None:
      sys.stderr.write("Problem %s undeclared\n" % problemName)
      sys.stderr.flush()
      return
    if (algorithmName=="FindBestSolution" or algorithmName=="BranchAndBound") and self._problem.GetObjectivePolynomial()==None:
      sys.stderr.write("Problem %s is not an optimization problem, change algorithm to 'FindOneSolution'\n" % problemName)
      sys.stderr.flush()
      algorithmName = "FindOneSolution"
    startTime = time.time()
    if algorithmName=="GenerateAndTest":
      solutions = self._solver.GenerateAndTest(self._problem)
    elif algorithmName=="FindOneSolution":
      solutions = self._solver.FindOneSolution(self._problem)
    elif algorithmName=="FindAllSolutions":
      solutions = self._solver.FindAllSolutions(self._problem)
    elif algorithmName=="FindBestSolution":
      solutions = self._solver.FindBestSolution(self._problem)
    elif algorithmName=="BranchAndBound":
      solutions = self._solver.BranchAndBound(self._problem)
    else:
      sys.stderr.write("Unknown algorithm %s\n" % algorithmName)
      sys.stderr.flush()
      return
    endTime = time.time()
    calculationTime = endTime - startTime
    if outputFileName==None:
      outputFile = sys.stdout
    else:
      try:
        outputFile = open(outputFileName, "w")
      except:
        sys.stderr.write("Unable to iopen file %s, printing results on the standard output\n" % outputFileName)
        sys.stderr.flush()
        outputFile = sys.stdout
    for solution in solutions:
      self._printer.PrintAssignment(self._problem, solution, outputFile)
    outputFile.write("%d solution(s)\n" % len(solutions))
    outputFile.write("%d choice(s)\n" % self._solver.numberOfChoices)
    outputFile.write("Execution time: %.2g second(s)\n" % calculationTime)
    if (algorithmName=="FindBestSolution" or algorithmName=="BranchAndBound") and len(solutions)>0:
      outputFile.write("Best score: %d\n" % self._solver._bestScore)
    outputFile.flush()
    if outputFile!=sys.stdout:
      outputFile.close()

