# Sherlock version 1.0.0
# SherlockParser.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Sherlock, a primitive solver for numerical contraint satisfaction problems.
# Sherlock is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Sherlock is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Sherlock.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Parser

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from SherlockModels import *
from SherlockLexer import *

# 2) Parser
# ---------

class SherlockParser:
  def __init__(self):
    self._lexer = SherlockLexer()
    self._errors = 0
    self._errorFile = sys.stderr
    self._currentTokenIndex = 0
    self._model = None
    self._problem = None

  def ImportFile(self, model, fileName):
    try:
      file = open(fileName, "r")
    except:
      sys.stderr.write('Unable to open file "%s"\n' % fileName)
      sys.stderr.flush()
      return 1
    self.ParseFile(model, fileName, file)
    file.close()
    return self._errors

  def ParseFile(self, model, fileName, file):
    self._errors = self._lexer.Initialize(fileName, file)
    if self._errors!=0:
      return
    self._ParseModel(model)
    self._lexer.Finalize()
    return self._errors

  def _ParseModel(self, model):
    self._currentTokenIndex = 0
    self._model = model
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None:
        break
      elif token.type==SherlockToken.KEYWORD_DOMAIN:
        domain = self._ParseDomainDeclaration()
        if domain==None:
          break
      elif token.type==SherlockToken.KEYWORD_PROBLEM:
        problem = self._ParseProblem()
        if problem==None:
          break
        problem.BuildOccurrenceLists()
      else:
        self._RaiseParseError(token, "A declaration of domain or problem was expected ")
        self._currentTokenIndex += 1

  def _ParseDomainDeclaration(self):
    token = self._ParseGivenToken(SherlockToken.KEYWORD_DOMAIN)
    if token==None:
      return None
    name = self._ParseIdentifier()
    if name==None:
      self._RaiseParseError(token, "An identifier was expected")
      return None
    domain = self._model.GetDomain(name)
    if domain!=None:
      self._RaiseParseError(token, "Domain %s already declared" % name)
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.type==SherlockToken.LBRACK:
      domain = self._ParseRange()
    elif token.type==SherlockToken.LBRACE:
      domain = self._ParseSet()
    else:
      self._RaiseParseError(token, "A domain was expected")
      return None
    if domain==None:
      return None
    domain.SetName(name)
    self._model.AddDomain(name, domain)
    return domain

  def _ParseRange(self):
    token = self._ParseGivenToken(SherlockToken.LBRACK)
    if token==None:
      return None
    lowerBound = self._ParseInteger()
    if lowerBound==None:
      return None
    token = self._ParseGivenToken(SherlockToken.COMMA)
    if token==None:
      return None
    upperBound = self._ParseInteger()
    if upperBound==None:
      return None
    token = self._ParseGivenToken(SherlockToken.RBRACK)
    if token==None:
      return None
    return SherlockDomain.NewRange(lowerBound, upperBound)

  def _ParseSet(self):
    token = self._ParseGivenToken(SherlockToken.LBRACE)
    if token==None:
      return None
    values = []
    while True:
      value = self._ParseInteger()
      values.append(value)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.type!=SherlockToken.COMMA:
        break
      self._currentTokenIndex += 1
    token = self._ParseGivenToken(SherlockToken.RBRACE)
    if token==None:
      return None
    return SherlockDomain.NewSet(values)

  def _ParseProblem(self):
    token = self._ParseGivenToken(SherlockToken.KEYWORD_PROBLEM)
    if token==None:
      return None
    name = self._ParseIdentifier()
    if name==None:
      self._RaiseParseError(token, "An identifier was expected")
      return None
    self._problem = self._model.GetProblem(name)
    if self._problem!=None:
      self._RaiseParseError(token, "Problem %s already declared" % name)
      return None
    self._problem = SherlockProblem(name)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None:
        self._RaiseParseError(token, "Unexpected end of file")
        return None
      if token.type==SherlockToken.KEYWORD_CONSTRAINT:
        ok = self._ParseConstraintClause()
        if not ok:
          return None
      elif token.type==SherlockToken.KEYWORD_MAXIMIZE or token.type==SherlockToken.KEYWORD_MINIMIZE:
        ok = self._ParseObjectiveFunction()
        if not ok:
          return None
      elif token.type==SherlockToken.KEYWORD_END:
        break
      else:
        variable = self._ParseVariableDeclaration()
        if variable==None:
          return None
    token = self._ParseGivenToken(SherlockToken.KEYWORD_END)
    if token==None:
      return None
    self._model.AddProblem(name, self._problem)
    return self._problem

  def _ParseVariableDeclaration(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.type==SherlockToken.LBRACK:
      domain = self._ParseRange()
      if domain==None:
        return None
    elif token.type==SherlockToken.LBRACE:
      domain = self._ParseSet()
      if domain==None:
        return None
    else:
      domainName = self._ParseIdentifier()
      if domainName==None:
        return None
      domain = self._model.GetDomain(domainName)
      if domain==None:
        self._RaiseParseError(token, "Undeclared domain %s" % domainName)
        return None
    variables = []
    while True:
      name = self._ParseIdentifier()
      if name==None:
        return None
      variable = self._problem.GetVariable(name)
      if variable!=None:
        self._RaiseParseError(token, "Variable %s already declared" % name)
        return None
      variable = self._problem.AddVariable(name, domain)
      variables.append(variable)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.type==SherlockToken.SEMICOLON:
        break;
      token = self._ParseGivenToken(SherlockToken.COMMA)
      if token==None:
        return None
    token = self._ParseGivenToken(SherlockToken.SEMICOLON)
    if token==None:
      return None
    return variables

  def _ParseConstraintClause(self):
    token = self._ParseGivenToken(SherlockToken.KEYWORD_CONSTRAINT)
    if token==None:
      return False
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None:
        self._RaiseParseError(token, "Unexpected end of file")
        return False
      if token.type==SherlockToken.KEYWORD_MAXIMIZE or token.type==SherlockToken.KEYWORD_MINIMIZE or token.type==SherlockToken.KEYWORD_END:
        break
      constraint = self._ParseConstraint()
      if constraint==None:
        return False
    return True

  def _ParseConstraint(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.type==SherlockToken.KEYWORD_ALLDIFFERENT:
      constraint = self._ParseAllDifferent()
    elif token.type==SherlockToken.KEYWORD_OR:
      constraint = self._ParseClause()
    else:
      constraint = self._ParseInequality()
    return constraint

  def _ParseAllDifferent(self):
    token = self._ParseGivenToken(SherlockToken.KEYWORD_ALLDIFFERENT)
    if token==None:
      return None
    variables = []
    while True:
      variable = self._ParseVariable()
      if variable==None:
        return None
      variables.append(variable)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.type==SherlockToken.SEMICOLON:
        break;
      token = self._ParseGivenToken(SherlockToken.COMMA)
      if token==None:
        return None
    token = self._ParseGivenToken(SherlockToken.SEMICOLON)
    if token==None:
      return None
    if len(variables)==0:
      token = self._lexer.GetToken(self._currentTokenIndex)
      self._RaiseParseError(token, "No variable in the constraint alldifferent")
    return self._problem.AddAllDifferent(variables)

  def _ParseClause(self):
    token = self._ParseGivenToken(SherlockToken.KEYWORD_OR)
    if token==None:
      return None
    literals = []
    while True:
      literal = self._ParseLiteral()
      if literal==None:
        return None
      literals.append(literal)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.type==SherlockToken.SEMICOLON:
        break;
      token = self._ParseGivenToken(SherlockToken.COMMA)
      if token==None:
        return None
    token = self._ParseGivenToken(SherlockToken.SEMICOLON)
    if token==None:
      return None
    if len(literals)==0:
      token = self._lexer.GetToken(self._currentTokenIndex)
      self._RaiseParseError(token, "Empty clause")
    return self._problem.AddClause(literals)

  def _ParseLiteral(self):
    variable = self._ParseVariable()
    if variable==None:
      return None
    comparator = self._ParseComparator()
    if comparator==None:
      return None
    value = self._ParseInteger()
    if value==None:
      return None
    if comparator.type==SherlockToken.EQ:
      literal = SherlockLiteral(SherlockLiteral.EQUAL, variable, value) 
    elif comparator.type==SherlockToken.DF:
      literal = SherlockLiteral(SherlockLiteral.DIFFERENT, variable, value) 
    elif comparator.type==SherlockToken.LT:
      literal = SherlockLiteral(SherlockLiteral.LESS, variable, value) 
    elif comparator.type==SherlockToken.GT:
      literal = SherlockLiteral(SherlockLiteral.GREATER, variable, value) 
    elif comparator.type==SherlockToken.LEQ:
      literal = SherlockLiteral(SherlockLiteral.LESS_OR_EQUAL, variable, value) 
    elif comparator.type==SherlockToken.GEQ:
      literal = SherlockLiteral(SherlockLiteral.GREATER_OR_EQUAL, variable, value)
    else:
      literal = None
    return literal

  def _ParseInequality(self):
    polynomial = self._ParsePolynomial()
    if polynomial==None:
      return None
    comparator = self._ParseComparator()
    if comparator==None:
      return None
    bound = self._ParseInteger()
    if bound==None:
      return None
    token = self._ParseGivenToken(SherlockToken.SEMICOLON)
    if token==None:
      return None
    if comparator.type==SherlockToken.EQ:
      constraint = self._problem.AddInequality(SherlockConstraint.EQUAL, polynomial, bound) 
    elif comparator.type==SherlockToken.DF:
      constraint = self._problem.AddInequality(SherlockConstraint.DIFFERENT, polynomial, bound) 
    elif comparator.type==SherlockToken.LT:
      constraint = self._problem.AddInequality(SherlockConstraint.LESS, polynomial, bound) 
    elif comparator.type==SherlockToken.GT:
      constraint = self._problem.AddInequality(SherlockConstraint.GREATER, polynomial, bound) 
    elif comparator.type==SherlockToken.LEQ:
      constraint = self._problem.AddInequality(SherlockConstraint.LESS_OR_EQUAL, polynomial, bound) 
    elif comparator.type==SherlockToken.GEQ:
      constraint = self._problem.AddInequality(SherlockConstraint.GREATER_OR_EQUAL, polynomial, bound)
    else:
      constraint = None
    return constraint

  def _ParsePolynomial(self):
    polynomial = []
    while True:
      monomial = self._ParseMonomial()
      if monomial==None:
        return None
      polynomial.append(monomial)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.type==SherlockToken.MINUS:
        pass
      elif token.type==SherlockToken.PLUS:
        token = self._ParseGivenToken(SherlockToken.PLUS)
      else:
        break
    return polynomial

  def _ParseMonomial(self):
    sign = 1
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.type==SherlockToken.IDENTIFIER:
      coefficient = 1
    elif token.type==SherlockToken.MINUS:
      sign = -1
      token = self._ParseGivenToken(SherlockToken.MINUS)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.type==SherlockToken.IDENTIFIER:
        coefficient = 1
      else:
        coefficient = self._ParseInteger()
        if coefficient==None:
          return None
        token = self._ParseGivenToken(SherlockToken.STAR)
        if token==None:
          return None
    else:
      coefficient = self._ParseInteger()
      if coefficient==None:
        return None
      token = self._ParseGivenToken(SherlockToken.STAR)
      if token==None:
        return None
    variable = self._ParseVariable()
    if variable==None:
      return None
    return SherlockMonomial(sign*coefficient, variable)

  def _ParseVariable(self):
    name = self._ParseIdentifier()
    if name==None:
      return None
    variable = self._problem.GetVariable(name)
    if variable==None:
      token = self._lexer.GetToken(self._currentTokenIndex)
      self._RaiseParseError(token, "Undeclared variable %s" % name)
      return None
    return variable

  def _ParseComparator(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if not self._IsComparator(token):
      self._RaiseParseError(token, "A comparator was expected")
      return None
    self._currentTokenIndex += 1
    return token

  def _ParseObjectiveFunction(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.type==SherlockToken.KEYWORD_MAXIMIZE:
      isMinimizationProblem = False
    elif token.type==SherlockToken.KEYWORD_MINIMIZE:
      isMinimizationProblem = True
    else:
      self._RaiseParseError(token, "Keyword minimize or maximize was expected")
      return False
    self._currentTokenIndex += 1
    polynomial = self._ParsePolynomial()
    if polynomial==None:
      return False
    token = self._ParseGivenToken(SherlockToken.SEMICOLON)
    if token==None:
      return False
    self._problem.SetObjectivePolynomial(polynomial, isMinimizationProblem)
    return True

  def _IsComparator(self, token):
    return token.type==SherlockToken.EQ or token.type==SherlockToken.DF or token.type==SherlockToken.LT or token.type==SherlockToken.GT or token.type==SherlockToken.LEQ or token.type==SherlockToken.GEQ

  def _ParseIdentifier(self):
    token = self._ParseGivenToken(SherlockToken.IDENTIFIER)
    if token==None:
      return None
    return token.string

  def _ParseInteger(self):
    sign = 1
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.type==SherlockToken.MINUS:
      sign = -1
      self._currentTokenIndex += 1
    token = self._ParseGivenToken(SherlockToken.INTEGER)
    if token==None:
      return None
    return sign*int(token.string)

  def _ParseGivenToken(self, tokenType):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None or token.type!=tokenType:
      self._RaiseParseError(token, "A %s was expected" % SherlockToken.GetNameOfType(tokenType))
      return None
    self._currentTokenIndex += 1
    return token

  def GetErrorFile(self):
    return self._errorFile
  def SetErrorFile(self, file):
    self._errorFile = file

  def _RaiseParseError(self, token, message):
    self._errors += 1
    if token!=None:
      self._errorFile.write("Parse error line %d (%d, %s):\n" % (token.lineNumber, token.columnNumber, token.string))
    self._errorFile.write("%s\n\n" % message)
    self._errorFile.flush()
